@@ -20,8 +20,15 @@ OpenStack Client interface. Handles the REST calls and responses.
 
 from __future__ import print_function
 
+import glob
+import hashlib
+import imp
+import itertools
 import logging
+import os
+import pkgutil
 import re
+import six
 
 from keystoneclient import access
 from keystoneclient import adapter
@@ -30,8 +37,11 @@ from keystoneclient import discover
 import requests
 
 from cinderclient import exceptions
+import cinderclient.extension
 from cinderclient.openstack.common import importutils
-from cinderclient.openstack.common import strutils
+from cinderclient.openstack.common.gettextutils import _
+from oslo_utils import encodeutils
+from oslo_utils import strutils
 
 osprofiler_web = importutils.try_import("osprofiler.web")
 
@@ -72,8 +82,8 @@ def get_volume_api_from_url(url):
         if version in components:
             return version[1:]
 
-    msg = "Invalid client version '%s'. must be one of: %s" % (
-        (version, ', '.join(_VALID_VERSIONS)))
+    msg = (_("Invalid url: '%(url)s'. It must include one of: %(version)s.")
+        % {'url': url, 'version': ', '.join(_VALID_VERSIONS)})
     raise exceptions.UnsupportedVersion(msg)
 
 
@@ -110,7 +120,14 @@ class SessionClient(adapter.LegacyJsonAdapter):
         return self._cs_request(url, 'DELETE', **kwargs)
 
     def get_volume_api_version_from_endpoint(self):
-        return get_volume_api_from_url(self.get_endpoint())
+        try:
+            version = get_volume_api_from_url(self.get_endpoint())
+        except exceptions.UnsupportedVersion as e:
+            msg = (_("Service catalog returned invalid url.\n"
+                     "%s") % six.text_type(e.message))
+            raise exceptions.UnsupportedVersion(msg)
+
+        return version
 
     def authenticate(self, auth=None):
         self.invalidate(auth)
@@ -130,6 +147,7 @@ class SessionClient(adapter.LegacyJsonAdapter):
 
 class HTTPClient(object):
 
+    SENSITIVE_HEADERS = ('X-Auth-Token', 'X-Subject-Token',)
     USER_AGENT = 'python-cinderclient'
 
     def __init__(self, user, password, projectid, auth_url=None,
@@ -183,6 +201,16 @@ class HTTPClient(object):
 
         self._logger = logging.getLogger(__name__)
 
+    def _safe_header(self, name, value):
+        if name in HTTPClient.SENSITIVE_HEADERS:
+            encoded = value.encode('utf-8')
+            hashed = hashlib.sha1(encoded)
+            digested = hashed.hexdigest()
+            return encodeutils.safe_decode(name), "{SHA1}%s" % digested
+        else:
+            return (encodeutils.safe_decode(name),
+                    encodeutils.safe_decode(value))
+
     def http_log_req(self, args, kwargs):
         if not self.http_log_debug:
             return
@@ -195,7 +223,8 @@ class HTTPClient(object):
                 string_parts.append(' %s' % element)
 
         for element in kwargs['headers']:
-            header = ' -H "%s: %s"' % (element, kwargs['headers'][element])
+            header = ("-H '%s: %s'" %
+                      self._safe_header(element, kwargs['headers'][element]))
             string_parts.append(header)
 
         if 'data' in kwargs:
@@ -312,7 +341,20 @@ class HTTPClient(object):
         return self._cs_request(url, 'DELETE', **kwargs)
 
     def get_volume_api_version_from_endpoint(self):
-        return get_volume_api_from_url(self.management_url)
+        try:
+            version = get_volume_api_from_url(self.management_url)
+        except exceptions.UnsupportedVersion as e:
+            if self.management_url == self.bypass_url:
+                msg = (_("Invalid url was specified in --bypass-url or "
+                         "environment variable CINDERCLIENT_BYPASS_URL.\n"
+                         "%s") % six.text_type(e.message))
+            else:
+                msg = (_("Service catalog returned invalid url.\n"
+                         "%s") % six.text_type(e.message))
+
+            raise exceptions.UnsupportedVersion(msg)
+
+        return version
 
     def _extract_service_catalog(self, url, resp, body, extract_token=True):
         """See what the auth service told us and process the response.
@@ -332,7 +374,8 @@ class HTTPClient(object):
                 management_url = self.service_catalog.url_for(
                     region_name=self.region_name,
                     endpoint_type=self.endpoint_type,
-                    service_type=self.service_type)
+                    service_type=self.service_type,
+                    service_name=self.service_name)
                 self.management_url = management_url.rstrip('/')
                 return None
             except exceptions.AmbiguousEndpoints:
@@ -547,6 +590,45 @@ def get_client_class(version):
     return importutils.import_class(client_path)
 
 
+def discover_extensions(version):
+    extensions = []
+    for name, module in itertools.chain(
+            _discover_via_python_path(),
+            _discover_via_contrib_path(version)):
+
+        extension = cinderclient.extension.Extension(name, module)
+        extensions.append(extension)
+
+    return extensions
+
+
+def _discover_via_python_path():
+    for (module_loader, name, ispkg) in pkgutil.iter_modules():
+        if name.endswith('python_cinderclient_ext'):
+            if not hasattr(module_loader, 'load_module'):
+                # Python 2.6 compat: actually get an ImpImporter obj
+                module_loader = module_loader.find_module(name)
+
+            module = module_loader.load_module(name)
+            yield name, module
+
+
+def _discover_via_contrib_path(version):
+    module_path = os.path.dirname(os.path.abspath(__file__))
+    version_str = "v%s" % version.replace('.', '_')
+    ext_path = os.path.join(module_path, version_str, 'contrib')
+    ext_glob = os.path.join(ext_path, "*.py")
+
+    for ext_path in glob.iglob(ext_glob):
+        name = os.path.basename(ext_path)[:-3]
+
+        if name == "__init__":
+            continue
+
+        module = imp.load_source(name, ext_path)
+        yield name, module
+
+
 def Client(version, *args, **kwargs):
     client_class = get_client_class(version)
     return client_class(*args, **kwargs)
